/** Rill v0.0.504 https://www.npmjs.com/package/rill License: MIT*/
var Application, Request, Response, co, compose, expect, http, isType, pathToReg, respond,
  slice = [].slice;

co = require("co");

pathToReg = require("path-to-regexp");

isType = require("is-type");

compose = require("./util/compose");

expect = require("./util/expect");

Request = require("./request");

Response = require("./response");

respond = require("./util/response");

http = (require("./hijackBrowser"));


/*
 * Creates an isomorphic Rill application that can listen for browser and server page requests.
#
 * @constructor
 */

Application = (function() {
  var j, len, method, ref;

  function Application() {
    if (!(this instanceof Application)) {
      return new Application();
    }
    this.middleware = [];
  }


  /*
  	 * Starts the server or browser hijacker.
   */

  Application.prototype.listen = function() {
    var server;
    server = http.createServer(this.callback());
    return server.listen.apply(server, arguments);
  };


  /*
  	 * Creates a call back function for the current app that can handle incoming requests.
   */

  Application.prototype.callback = function() {
    var fn;
    fn = co.wrap(compose([respond].concat(this.middleware)));
    return function(req, res) {
      if (res == null) {
        res = {};
      }
      res.statusCode = 404;
      req = new Request(req);
      res = new Response(res);
      return fn.call({}, req, res).then(function() {
        return {
          redirected: Boolean(req.redirectURL)
        };
      })["catch"](function(err) {
        return console.error(err.stack || err);
      });
    };
  };


  /*
  	 * Attaches a midleware function to the current app that will be ran with requests.
  	 * @param {Generator} handlers
   */

  Application.prototype.use = function() {
    var handler, handlers, j, len, ref;
    handlers = 1 <= arguments.length ? slice.call(arguments, 0) : [];
    for (j = 0, len = handlers.length; j < len; j++) {
      handler = handlers[j];
      expect((handler != null ? (ref = handler.constructor) != null ? ref.name : void 0 : void 0) === "GeneratorFunction", "Can only use generator functions.");
      this.middleware.push(handler);
    }
    return this;
  };


  /*
  	 * Simple syntactic sugar for functions that wish to modify the current application instance.
  	#
  	 * @param {Function} transformers
   */

  Application.prototype.setup = function() {
    var j, len, transformer, transformers;
    transformers = 1 <= arguments.length ? slice.call(arguments, 0) : [];
    for (j = 0, len = transformers.length; j < len; j++) {
      transformer = transformers[j];
      expect(transformer(this) === this, "Setup failed, transformer function must return the modified application.");
    }
    return this;
  };


  /*
  	 * Attaches a seperate rill application at a given path.
  	 * req.path will be updated to remove the provided path.
  	#
  	 * @param {Application} apps
   */

  Application.prototype.mount = function() {
    var app, apps, downstream, j, len, path;
    path = arguments[0], apps = 2 <= arguments.length ? slice.call(arguments, 1) : [];
    if (typeof path !== "string") {
      apps = path;
      path = "/";
    }
    expect(path[0] === "/", "Mount path must begin with '/'.");
    downstream = [];
    for (j = 0, len = apps.length; j < len; j++) {
      app = apps[j];
      expect(app instanceof Application, "Can only mount other Rill apps.");
      downstream = downstream.concat(app.middleware);
    }
    downstream = compose(downstream);
    if (path === "/") {
      this.use(downstream);
      return this;
    }
    path = pathToReg(path, [], {
      end: false
    });
    this.use(regeneratorRuntime.mark(function callee$2$0(req, res, upstream) {
      var originalPath;

      return regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
        while (1) switch (context$3$0.prev = context$3$0.next) {
        case 0:
          if (!path.test(req.path)) {
            context$3$0.next = 7;
            break;
          }

          originalPath = req.path;
          context$3$0.next = 4;

          return downstream.call(this, req, res, (regeneratorRuntime.mark(function callee$3$0() {
            return regeneratorRuntime.wrap(function callee$3$0$(context$4$0) {
              while (1) switch (context$4$0.prev = context$4$0.next) {
              case 0:
                context$4$0.next = 2;
                return upstream;
              case 2:
                return context$4$0.abrupt("return", req.path = req.path.replace(path, ""));
              case 3:
              case "end":
                return context$4$0.stop();
              }
            }, callee$3$0, this);
          }))());
        case 4:
          return context$3$0.abrupt("return", req.path = originalPath);
        case 7:
          context$3$0.next = 9;
          return upstream;
        case 9:
          return context$3$0.abrupt("return", context$3$0.sent);
        case 10:
        case "end":
          return context$3$0.stop();
        }
      }, callee$2$0, this);
    }));
    return this;
  };


  /*
  	 * Create functions for each http verb that will accept a path url and a middleware function to use.
  	#
  	 * @param {String} path
  	 * @param {Generator} middleware
   */

  ref = ["options", "head", "get", "put", "post", "del"];
  for (j = 0, len = ref.length; j < len; j++) {
    method = ref[j];
    Application.prototype[method] = (function(method) {
      return function() {
        var keys, middleware, middlewares, path;
        path = arguments[0], middlewares = 2 <= arguments.length ? slice.call(arguments, 1) : [];
        keys = [];
        method = method.toUpperCase();
        middleware = compose(middlewares);
        if (!isType.RegExp(path)) {
          path = pathToReg(path, keys);
        }
        this.use(regeneratorRuntime.mark(function callee$3$0(req, res, next) {
          var i, k, len1, match, name, optional, ref1, val;

          return regeneratorRuntime.wrap(function callee$3$0$(context$4$0) {
            while (1) switch (context$4$0.prev = context$4$0.next) {
            case 0:
              if (!(req.method === method && (match = req.path.match(path)))) {
                context$4$0.next = 19;
                break;
              }

              req.params = {};
              i = k = 0, len1 = keys.length;
            case 3:
              if (!(k < len1)) {
                context$4$0.next = 14;
                break;
              }

              ref1 = keys[i], name = ref1.name, optional = ref1.optional;
              val = match[i + 1];

              if (optional || (val != null)) {
                context$4$0.next = 10;
                break;
              }

              context$4$0.next = 9;
              return next;
            case 9:
              return context$4$0.abrupt("return", context$4$0.sent);
            case 10:
              req.params[name] = val;
            case 11:
              i = ++k;
              context$4$0.next = 3;
              break;
            case 14:
              context$4$0.next = 16;
              return middleware.call(this, req, res, next);
            case 16:
              return context$4$0.abrupt("return", context$4$0.sent);
            case 19:
              context$4$0.next = 21;
              return next;
            case 21:
              return context$4$0.abrupt("return", context$4$0.sent);
            case 22:
            case "end":
              return context$4$0.stop();
            }
          }, callee$3$0, this);
        }));
        return this;
      };
    })(method);
  }

  return Application;

})();

module.exports = Application;
