co        = require("co")
pathToReg = require("path-to-regexp")
isType    = require("is-type")
compose   = require("./util/compose")
expect    = require("./util/expect")
Request   = require("./request")
Response  = require("./response")
respond   = require("./util/response")
http      = (
	if process.browser then require("./hijackBrowser")
	else http = require("http")
)

###
# Creates an isomorphic Rill application that can listen for browser and server page requests.
#
# @constructor
###
class Application
	constructor: ->
		return new Application() unless this instanceof Application
		@middleware = []

	###
	# Starts the server or browser hijacker.
	###
	listen: ->
		server = http.createServer(@callback())
		return server.listen.apply(server, arguments)

	###
	# Creates a call back function for the current app that can handle incoming requests.
	###
	callback: ->
		fn = co.wrap(compose([respond].concat(@middleware)))
		(req, res = {})->
			res.statusCode = 404
			req = new Request(req)
			res = new Response(res)

			fn.call({}, req, res)
			.then(-> redirected: Boolean(req.redirectURL))
			.catch((err)-> console.error(err.stack or err))

	###
	# Attaches a midleware function to the current app that will be ran with requests.
	# @param {Generator} handlers
	###
	use: (handlers...)->
		for handler in handlers
			expect(handler?.constructor?.name is "GeneratorFunction", "Can only use generator functions.")
			@middleware.push(handler)
		return this

	###
	# Simple syntactic sugar for functions that wish to modify the current application instance.
	#
	# @param {Function} transformers
	###
	setup: (transformers...)->
		for transformer in transformers
			expect(transformer(@) is @, "Setup failed, transformer function must return the modified application.")
		return this

	###
	# Attaches a seperate rill application at a given path.
	# req.path will be updated to remove the provided path.
	#
	# @param {Application} apps
	###
	mount: (path, apps...)->
		# Make path optional.
		unless typeof path is "string"
			apps = path
			path = "/"
		expect(path[0] is "/", "Mount path must begin with '/'.")
		downstream = []

		for app in apps
			expect(app instanceof Application, "Can only mount other Rill apps.")
			downstream = downstream.concat(app.middleware)

		downstream = compose(downstream)

		# Skip mounting.
		if path is "/"
			@use(downstream)
			return this

		path = pathToReg(path, [], end: no)

		@use((req, res, upstream)->
			if path.test(req.path)
				originalPath = req.path
				yield downstream.call(@, req, res, do ->
					# Wait for previous middleware to finish before modifying path.
					yield upstream
					req.path = req.path.replace(path, "")
				)
				req.path = originalPath
			else yield upstream
		)
		return this

	###
	# Create functions for each http verb that will accept a path url and a middleware function to use.
	#
	# @param {String} path
	# @param {Generator} middleware
	###
	for method in ["options", "head", "get", "put", "post", "del"]
		@::[method] = do (method)-> (path, middlewares...)->
			keys       = []
			method     = method.toUpperCase()
			middleware = compose(middlewares)
			path       = pathToReg(path, keys) unless isType.RegExp(path)

			@use((req, res, next)->
				if (req.method is method) and match = req.path.match(path)
					req.params = {}
					for { name, optional }, i in keys
						val = match[i + 1]
						return yield next unless optional or val?
						req.params[name] = val
					yield middleware.call(@, req, res, next)
				else yield next
			)
			return this


module.exports = Application
